@@ -2,41 +2,43 @@ package svu
 
 import (
 	"fmt"
+	"log"
 	"regexp"
 	"strconv"
 	"strings"
 
 	"github.com/Masterminds/semver/v3"
-	"github.com/caarlos0/svu/v2/internal/git"
+	"github.com/caarlos0/svu/v3/internal/git"
 )
 
+type Action uint
+
 const (
-	NextCmd       = "next"
-	MajorCmd      = "major"
-	MinorCmd      = "minor"
-	PatchCmd      = "patch"
-	CurrentCmd    = "current"
-	PreReleaseCmd = "prerelease"
+	Next Action = iota
+	Major
+	Minor
+	Patch
+	Current
+	PreRelease
 )
 
 var (
-	breaking     = regexp.MustCompile("(?m).*BREAKING[ -]CHANGE:.*")
-	breakingBang = regexp.MustCompile(`(?im).*(\w+)(\(.*\))?!:.*`)
+	breakingBody = regexp.MustCompile("(?m).*BREAKING[ -]CHANGE:.*")
+	breaking     = regexp.MustCompile(`(?im).*(\w+)(\(.*\))?!:.*`)
 	feature      = regexp.MustCompile(`(?im).*feat(\(.*\))?:.*`)
 	patch        = regexp.MustCompile(`(?im).*fix(\(.*\))?:.*`)
 )
 
 type Options struct {
-	Cmd                       string
-	Pattern                   string
-	Prefix                    string
-	StripPrefix               bool
-	PreRelease                string
-	Build                     string
-	Directory                 string
-	TagMode                   string
-	ForcePatchIncrement       bool
-	PreventMajorIncrementOnV0 bool
+	Action      Action
+	Pattern     string
+	Prefix      string
+	PreRelease  string
+	Metadata    string
+	Directories []string
+	TagMode     string
+	Always      bool
+	KeepV0      bool
 }
 
 func Version(opts Options) (string, error) {
@@ -50,101 +52,87 @@ func Version(opts Options) (string, error) {
 		return "", fmt.Errorf("could not get current version from tag: '%s': %w", tag, err)
 	}
 
-	result, err := nextVersion(
-		string(opts.Cmd),
-		current,
-		tag,
-		opts.PreRelease,
-		opts.Build,
-		opts.Directory,
-		opts.PreventMajorIncrementOnV0,
-		opts.ForcePatchIncrement,
-	)
+	result, err := nextVersion(current, tag, opts)
 	if err != nil {
 		return "", fmt.Errorf("could not get next tag: '%s': %w", tag, err)
 	}
 
-	if opts.StripPrefix {
-		return result.String(), nil
-	}
 	return opts.Prefix + result.String(), nil
 }
 
-func nextVersion(cmd string, current *semver.Version, tag, preRelease, build, directory string, preventMajorIncrementOnV0, forcePatchIncrement bool) (semver.Version, error) {
-	if cmd == CurrentCmd {
+func nextVersion(
+	current *semver.Version,
+	tag string,
+	opts Options,
+) (semver.Version, error) {
+	if opts.Action == Current {
 		return *current, nil
 	}
 
-	if forcePatchIncrement {
-		c, err := current.SetMetadata("")
-		if err != nil {
-			return c, err
-		}
-		c, err = c.SetPrerelease("")
-		if err != nil {
-			return c, err
-		}
+	if opts.Always {
+		c, _ := current.SetMetadata("")
+		c, _ = c.SetPrerelease("")
 		current = &c
 	}
 
 	var result semver.Version
 	var err error
-	switch cmd {
-	case NextCmd, PreReleaseCmd:
-		result, err = findNextWithGitLog(current, tag, directory, preventMajorIncrementOnV0, forcePatchIncrement)
-	case MajorCmd:
+	switch opts.Action {
+	case Next, PreRelease:
+		result, err = findNextWithGitLog(current, tag, opts)
+	case Major:
 		result = current.IncMajor()
-	case MinorCmd:
+	case Minor:
 		result = current.IncMinor()
-	case PatchCmd:
+	case Patch:
 		result = current.IncPatch()
 	}
 	if err != nil {
 		return result, err
 	}
 
-	if cmd == PreReleaseCmd {
-		result, err = nextPreRelease(current, &result, preRelease)
+	if opts.Action == PreRelease {
+		result, err = nextPreRelease(current, &result, opts.PreRelease)
 		if err != nil {
 			return result, err
 		}
 	} else {
-		result, err = result.SetPrerelease(preRelease)
+		result, err = result.SetPrerelease(opts.PreRelease)
 		if err != nil {
 			return result, err
 		}
 	}
 
-	result, err = result.SetMetadata(build)
+	result, err = result.SetMetadata(opts.Metadata)
 	if err != nil {
 		return result, err
 	}
 	return result, nil
 }
 
-func nextPreRelease(current, next *semver.Version, preRelease string) (semver.Version, error) {
+func nextPreRelease(current, next *semver.Version, prerelease string) (semver.Version, error) {
 	suffix := ""
-	if preRelease != "" {
+	if prerelease != "" {
 		// Check if the suffix already contains a version number, if it does assume the user wants to explicitly set the version so use that
-		splitPreRelease := strings.Split(preRelease, ".")
+		splitPreRelease := strings.Split(prerelease, ".")
 		if len(splitPreRelease) > 1 {
 			if _, err := strconv.Atoi(splitPreRelease[len(splitPreRelease)-1]); err == nil {
-				return current.SetPrerelease(preRelease)
+				return current.SetPrerelease(prerelease)
 			}
 		}
 
-		suffix = preRelease
+		suffix = prerelease
 
 		// Check if the prerelease suffix is the same as the current prerelease
 		preSuffix := strings.Split(current.Prerelease(), ".")[0]
-		if preSuffix == preRelease {
+		if preSuffix == prerelease {
 			suffix = current.Prerelease()
 		}
 	} else if current.Prerelease() != "" {
 		suffix = current.Prerelease()
 	} else {
 		return *current, fmt.Errorf(
-			"--pre-release suffix is required to calculate next pre-release version as suffix could not be determined from current version: %s",
+			"--prerelease suffix is required to calculate next pre-release version as suffix could not be determined from current version: %s",
 			current.String(),
 		)
 	}
@@ -181,42 +169,70 @@ func getCurrentVersion(tag, prefix string) (*semver.Version, error) {
 	return current, err
 }
 
-func findNextWithGitLog(current *semver.Version, tag string, directory string, preventMajorIncrementOnV0, forcePatchIncrement bool) (semver.Version, error) {
-	log, err := git.Changelog(tag, directory)
+func findNextWithGitLog(
+	current *semver.Version,
+	tag string,
+	opts Options,
+) (semver.Version, error) {
+	log, err := git.Changelog(tag, opts.Directories)
 	if err != nil {
 		return semver.Version{}, fmt.Errorf("failed to get changelog: %w", err)
 	}
 
-	return findNext(current, preventMajorIncrementOnV0, forcePatchIncrement, log), nil
+	return findNext(current, log, opts), nil
 }
 
-func isBreaking(log string) bool {
-	return breaking.MatchString(log) || breakingBang.MatchString(log)
+func isBreaking(commit git.Commit) bool {
+	return breakingBody.MatchString(commit.Body) || breaking.MatchString(commit.Title)
 }
 
-func isFeature(log string) bool {
-	return feature.MatchString(log)
+func isFeature(commit git.Commit) bool {
+	return feature.MatchString(commit.Title)
 }
 
-func isPatch(log string) bool {
-	return patch.MatchString(log)
+func isPatch(commit git.Commit) bool {
+	return patch.MatchString(commit.Title)
 }
 
-func findNext(current *semver.Version, preventMajorIncrementOnV0, forcePatchIncrement bool, log string) semver.Version {
-	if isBreaking(log) {
-		if current.Major() == 0 && preventMajorIncrementOnV0 {
+func findNext(current *semver.Version, changes []git.Commit, opts Options) semver.Version {
+	var major, minor, patch *git.Commit
+	for _, commit := range changes {
+		if isBreaking(commit) {
+			major = &commit
+			break // no bigger change allowed, so we're done
+		}
+
+		if minor == nil && isFeature(commit) {
+			minor = &commit
+		}
+
+		if patch == nil && isPatch(commit) {
+			patch = &commit
+		}
+	}
+
+	if major != nil {
+		if current.Major() == 0 && opts.KeepV0 {
+			log.Printf("found major change, but 'keep v0' is set: %s %s\n", major.SHA, major.Title)
 			return current.IncMinor()
 		}
+		log.Printf("found major change: %s %s\n", major.SHA, major.Title)
 		return current.IncMajor()
 	}
 
-	if isFeature(log) {
+	if minor != nil {
+		log.Printf("found minor change: %s %s\n", minor.SHA, minor.Title)
 		return current.IncMinor()
 	}
 
-	if forcePatchIncrement || isPatch(log) {
+	if patch != nil {
+		log.Printf("found patch change: %s %s\n", patch.SHA, patch.Title)
 		return current.IncPatch()
 	}
 
+	if opts.Always {
+		log.Printf("found no changes, but 'always' is set")
+		return current.IncPatch()
+	}
 	return *current
 }
